"""RadioBrowser API Struct Tag"""

from msgspec import Struct

from typing import Optional


class Tag(Struct):
    """RadioBrowser API Struct Tag

    Attributes
    ----------------------------------------
    ``name`` : |str|, optional
        Name of the tag

    ``stationcount`` : |int|, optional
        Station tag count
    """
    name: Optional[str] = None
    stationcount: Optional[int] = None


"""RadioBrowser API Struct Station Click"""

from typing import Optional, Union

from msgspec import Struct

from radiozilla.datatypes import UUID


class StationClick(Struct):
    """RadioBrowser API Struct Station Click

    Attributes:
    ----------------------------------------
    ``stationuuid`` : |str, URL|, optional
        An unique id for referencing a Station.

    ``clickuuid`` : |str, URL|, optional
        ...

    ``clicktimestamp_iso8601`` : |str, URL|, optional
        ...

    ``clicktimestamp`` : |str, URL|, optional
        ...
    """
    stationuuid: Optional[Union[str, UUID]] = None
    clickuuid: Optional[Union[str, UUID]] = None
    clicktimestamp_iso8601: Optional[Union[str, UUID]] = None
    clicktimestamp: Optional[Union[str, UUID]] = None



"""RadioBrowser API Struct CountryCode"""

from msgspec import Struct

from typing import Optional


class CountryCode(Struct):
    """RadioBrowser API Struct CountryCode

    Attributes
    ----------------------------------------
    ``name`` : |str|, optional
        Name of the countrycode

    ``stationcount`` : |int|, optional
        Station countrycode count
    """
    name: Optional[str] = None
    stationcount: Optional[int] = None


"""RadioBrowser API Struct Codec"""

from msgspec import Struct

from typing import Optional


class Codec(Struct):
    """RadioBrowser API Struct Codec

    Attributes
    ----------------------------------------
    ``name`` : |str|, optional
        Name of the codec

    ``stationcount`` : |int|, optional
        Station codec count
    """
    name: Optional[str] = None
    stationcount: Optional[int] = None


"""RadioBrowser API Struct Station Check Step"""

from msgspec import Struct

from typing import Optional, Union

from radiozilla.datatypes import UUID, URL


class StationCheckStep(Struct):
    """RadioBrowser API Struct Station Check Step

    Attributes:
    ----------------------------------------
    ``stepuuid`` : |str, UUID|, optional
        An unique id for this StationCheckStep.

    ``parent_stepuuid`` : |str, UUID|, optional
        An unique id for referencing another StationCheckStep.
        Is set if this step has a parent.

    ``checkuuid`` : |str, UUID|, optional
        An unique id for referencing a StationCheck.

    ``stationuuid`` : |str, UUID|, optional
        An unique id for referencing a Station.

    ``url`` : |str, URL|, optional
        The url that this step of the checking process handled.

    ``urltype`` : |str|, optional
        Does represent which kind of url it is. One of
        the following: STREAM, REDIRECT, PLAYLIST.

    ``error`` : |str|, optional
        URL to the homepage of the stream, so you can direct the
        user to a page with more information about the stream.

    ``creation_iso8601`` : |str|, optional
        Date and time of step creation.
    """

    stepuuid: Optional[Union[str, UUID]] = None
    parent_stepuuid: Optional[Union[str, UUID]] = None
    checkuuid: Optional[Union[str, UUID]] = None
    stationuuid: Optional[Union[str, UUID]] = None
    url: Optional[Union[str, URL]] = None
    urltype: Optional[str] = None
    error: Optional[str] = None
    creation_iso8601: Optional[str] = None


"""Main RadioBrowser API wrapper module"""

from asyncio import run
from typing import Any, Optional, Union

from aiohttp import ClientSession
from msgspec.json import decode

from . import __version__
from .decorators.memoizations import memoize
from .api.browsers import Browser
from .datatypes import T, Order, Boolean, NOT_SET
from .structs.codecs import Codec
from .structs.countries import Country
from .structs.countrycodes import CountryCode
from .structs.languages import Language
from .structs.states import State
from .structs.stations import Station
from .structs.tags import Tag
from .defaults import (
    _default_order,
    _default_reverse,
    _default_hidebroken,
    _default_offset,
    _default_limit,
    _default_name_exact,
    _default_country_exact,
    _default_language_exact,
    _default_has_extended_info,
    _default_has_geo_info,
    _default_is_https,
    _default_state_exact,
    _default_tag_exact,
)


class Radio(Browser):
    """Main RadioBrowser API wrapper class
    
    Attribtes:
    ----------------------------------------
    ``codecs`` : property
        Get list of codecs

    ``countries`` : property
        Get list of countries

    ``countrycodes`` : property
        Get list of countrycodes

    ``languages`` : property
        Get list of languages

    ``states`` : property
        Get list of states
    
    ``tags`` : property
        Get list of tags
    
    ``stations`` : property
        Get list of stations

    ``search`` : method
        Search station
    """

    def __init__(self) -> None:
        """"""
        super().__init__()

        self.headers = {"User-Agent": f"radiozilla/{__version__}"}

    @memoize
    async def __get(self, type: T, endpoint: str, params: dict[str, Any]) -> T:
        """Asynchronous method for sending requests to RadioBrowser API

        Parameters:
        ----------------------------------------
        ``type`` : |T|, required
            Serialization of the response with type

        ``endpoint`` : |str|, required
            Endpoint to request of RadioBrowser API

        ``params`` : |dict |str, Any||, optional
            Parameters of request
        """
        url = f"{self.url}/json/{endpoint}"

        async with ClientSession(headers=self.headers) as session:
            async with session.get(url, params=params) as response:
                return decode(await response.content.read(), type=type)

    def codecs(
        self,
        order: Optional[Union[str, Order]] = _default_order,
        reverse: Optional[Union[bool, Boolean]] = _default_reverse,
        offset: Optional[int] = _default_offset,
        limit: Optional[int] = _default_limit,
        hidebroken: Optional[Union[str, Boolean]] = _default_hidebroken,
    ) -> list[Codec]:
        """Get list of codecs

        Parameters:
        ----------------------------------------
        ``order`` : |str, Order|, optional
            Name of the attribute the result list will be sorted by.

        ``reverse`` : |str, Boolean|, optional
            Reverse the result list if set to true.

        ``offset`` : |str|, optional
            Starting value of the result list from the database.
            For example, if you want to do paging on the server side.

        ``limit`` : |str|, optional
            Number of returned datarows (stations) starting with offset.

        ``hidebroken`` : |str, Boolean|, optional
            Do list/not list broken stations.
        """
        params = {
            "order": order,
            "reverse": reverse,
            "offset": offset,
            "limit": limit,
            "hidebroken": hidebroken,
        }

        return run(self.__get(
            type=list[Country],
            endpoint="codecs",
            params=params
        ))

    def countries(
        self,
        order: Optional[Union[str, Order]] = _default_order,
        reverse: Optional[Union[bool, Boolean]] = _default_reverse,
        offset: Optional[int] = _default_offset,
        limit: Optional[int] = _default_limit,
        hidebroken: Optional[Union[str, Boolean]] = _default_hidebroken,
    ) -> list[Country]:
        """Get list of countries

        Parameters:
        ----------------------------------------
        ``order`` : |str, Order|, optional
            Name of the attribute the result list will be sorted by.

        ``reverse`` : |str, Boolean|, optional
            Reverse the result list if set to true.

        ``offset`` : |str|, optional
            Starting value of the result list from the database.
            For example, if you want to do paging on the server side.

        ``limit`` : |str|, optional
            Number of returned datarows (stations) starting with offset.

        ``hidebroken`` : |str, Boolean|, optional
            Do list/not list broken stations.
        """
        params = {
            "order": order,
            "reverse": reverse,
            "offset": offset,
            "limit": limit,
            "hidebroken": hidebroken,
        }

        return run(self.__get(
            type=list[Country],
            endpoint="countries",
            params=params
        ))

    def countrycodes(
        self,
        order: Optional[Union[str, Order]] = _default_order,
        reverse: Optional[Union[bool, Boolean]] = _default_reverse,
        offset: Optional[int] = _default_offset,
        limit: Optional[int] = _default_limit,
        hidebroken: Optional[Union[str, Boolean]] = _default_hidebroken,
    ) -> list[CountryCode]:
        """Get list of countrycodes

        Parameters:
        ----------------------------------------
        ``order`` : |str, Order|, optional
            Name of the attribute the result list will be sorted by.

        ``reverse`` : |str, Boolean|, optional
            Reverse the result list if set to true.

        ``offset`` : |str|, optional
            Starting value of the result list from the database.
            For example, if you want to do paging on the server side.

        ``limit`` : |str|, optional
            Number of returned datarows (stations) starting with offset.

        ``hidebroken`` : |str, Boolean|, optional
            Do list/not list broken stations.
        """
        params = {
            "order": order,
            "reverse": reverse,
            "offset": offset,
            "limit": limit,
            "hidebroken": hidebroken,
        }

        return run(self.__get(
            type=list[CountryCode],
            endpoint="countrycodes",
            params=params
        ))

    def languages(
        self,
        order: Optional[Union[str, Order]] = _default_order,
        reverse: Optional[Union[bool, Boolean]] = _default_reverse,
        offset: Optional[int] = _default_offset,
        limit: Optional[int] = _default_limit,
        hidebroken: Optional[Union[str, Boolean]] = _default_hidebroken,
    ) -> list[Language]:
        """Get list of languages

        Parameters:
        ----------------------------------------
        ``order`` : |str, Order|, optional
            Name of the attribute the result list will be sorted by.

        ``reverse`` : |str, Boolean|, optional
            Reverse the result list if set to true.

        ``offset`` : |str|, optional
            Starting value of the result list from the database.
            For example, if you want to do paging on the server side.

        ``limit`` : |str|, optional
            Number of returned datarows (stations) starting with offset.

        ``hidebroken`` : |str, Boolean|, optional
            Do list/not list broken stations.
        """
        params = {
            "order": order,
            "reverse": reverse,
            "offset": offset,
            "limit": limit,
            "hidebroken": hidebroken,
        }

        return run(self.__get(
            type=list[Language],
            endpoint="languages",
            params=params
        ))

    def states(
        self,
        order: Optional[Union[str, Order]] = _default_order,
        reverse: Optional[Union[bool, Boolean]] = _default_reverse,
        offset: Optional[int] = _default_offset,
        limit: Optional[int] = _default_limit,
        hidebroken: Optional[Union[str, Boolean]] = _default_hidebroken,
    ) -> list[State]:
        """Get list of states

        Parameters:
        ----------------------------------------
        ``order`` : |str, Order|, optional
            Name of the attribute the result list will be sorted by.

        ``reverse`` : |str, Boolean|, optional
            Reverse the result list if set to true.

        ``offset`` : |str|, optional
            Starting value of the result list from the database.
            For example, if you want to do paging on the server side.

        ``limit`` : |str|, optional
            Number of returned datarows (stations) starting with offset.

        ``hidebroken`` : |str, Boolean|, optional
            Do list/not list broken stations.
        """
        params = {
            "order": order,
            "reverse": reverse,
            "offset": offset,
            "limit": limit,
            "hidebroken": hidebroken,
        }

        return run(self.__get(
            type=list[State],
            endpoint="states",
            params=params
        ))

    def tags(
        self,
        order: Optional[Union[str, Order]] = _default_order,
        reverse: Optional[Union[bool, Boolean]] = _default_reverse,
        offset: Optional[int] = _default_offset,
        limit: Optional[int] = _default_limit,
        hidebroken: Optional[Union[str, Boolean]] = _default_hidebroken,
    ) -> list[Tag]:
        """Get list of tags

        Parameters:
        ----------------------------------------
        ``order`` : |str, Order|, optional
            Name of the attribute the result list will be sorted by.

        ``reverse`` : |str, Boolean|, optional
            Reverse the result list if set to true.

        ``offset`` : |str|, optional
            Starting value of the result list from the database.
            For example, if you want to do paging on the server side.

        ``limit`` : |str|, optional
            Number of returned datarows (stations) starting with offset.

        ``hidebroken`` : |str, Boolean|, optional
            Do list/not list broken stations.
        """
        params = {
            "order": order,
            "reverse": reverse,
            "offset": offset,
            "limit": limit,
            "hidebroken": hidebroken,
        }

        return run(self.__get(
            type=list[Tag],
            endpoint="tags",
            params=params
        ))

    def stations(
        self,
        order: Optional[Union[str, Order]] = _default_order,
        reverse: Optional[Union[bool, Boolean]] = _default_reverse,
        offset: Optional[int] = _default_offset,
        limit: Optional[int] = _default_limit,
        hidebroken: Optional[Union[str, Boolean]] = _default_hidebroken,
    ) -> list[Station]:
        """Get list of stations

        Parameters:
        ----------------------------------------
        ``order`` : |str, Order|, optional
            Name of the attribute the result list will be sorted by.

        ``reverse`` : |str, Boolean|, optional
            Reverse the result list if set to true.

        ``offset`` : |str|, optional
            Starting value of the result list from the database.
            For example, if you want to do paging on the server side.

        ``limit`` : |str|, optional
            Number of returned datarows (stations) starting with offset.

        ``hidebroken`` : |str, Boolean|, optional
            Do list/not list broken stations.
        """
        params = {
            "order": order,
            "reverse": reverse,
            "offset": offset,
            "limit": limit,
            "hidebroken": hidebroken,
        }

        return run(self.__get(
            type=list[Station],
            endpoint="stations",
            params=params
        ))

    def search(
        self,
        name: Optional[str] = NOT_SET,
        name_exact: Optional[Union[str, Boolean]] = _default_name_exact,
        country: Optional[str] = NOT_SET,
        country_exact: Optional[Union[str, Boolean]] = _default_country_exact,
        countrycode: Optional[str] = NOT_SET,
        state: Optional[str] = NOT_SET,
        state_exact: Optional[Union[str, Boolean]] = _default_state_exact,
        language: Optional[str] = NOT_SET,
        language_exact: Optional[Union[str, Boolean]] = _default_language_exact,
        tag: Optional[str] = NOT_SET,
        tag_exact: Optional[Union[str, Boolean]] = _default_tag_exact,
        tag_list: Optional[str] = NOT_SET,
        codec: Optional[str] = NOT_SET,
        bitrate_min: Optional[str] = NOT_SET,
        bitrate_max: Optional[str] = NOT_SET,
        has_geo_info: Optional[Union[str, Boolean]] = _default_has_geo_info,
        has_extended_info: Optional[Union[str, Boolean]] = _default_has_extended_info,
        is_https: Optional[Union[str, Boolean]] =_default_is_https,
        order: Optional[Union[str, Order]] = NOT_SET,
        reverse: Optional[Union[str, Boolean]] = _default_reverse,
        offset: Optional[str] = NOT_SET,
        limit: Optional[str] = NOT_SET,
        hidebroken: Optional[Union[str, Boolean]] = _default_hidebroken,
    ) -> list[Station]:
        """Search station

        Parameters:
        ----------------------------------------
        ``name`` : |str|, optional
            Name of the station.

        ``name_exact`` : |str, Boolean|, optional
            True: only exact matches, otherwise all matches.

        ``country`` : |str|, optional
            Country of the station.

        ``country_exact`` : |str, Boolean|, optional
            True: only exact matches, otherwise all matches.

        ``countrycode`` : |str|, optional
            2-digit countrycode of the station (see ISO 3166-1 alpha-2).

        ``state`` : |str|, optional
            State of the station

        ``state_exact`` : |str, Boolean|, optional
            True: only exact matches, otherwise all matches.

        ``language`` : |str|, optional
            Language of the station

        ``language_exact`` : |str, Boolean|, optional
            True: only exact matches, otherwise all matches.

        ``tag`` : |str|, optional
            A tag of the station

        ``tag_exact`` : |str, Boolean|, optional
            True: only exact matches, otherwise all matches.

        ``tag_list`` : |str|, optional
            A comma-separated list of tag. It can also be an
            array of string in JSON HTTP POST parameters. All
            tags in list have to match.

        ``codec`` : |str|, optional
            Codec of the station

        ``bitrate_min`` : |str|, optional
            Minimum of kbps for bitrate field of stations in result

        ``bitrate_max`` : |str|, optional
            Maximum of kbps for bitrate field of stations in result

        ``has_geo_info`` : |str, Boolean|, optional
            Not set = display all;
            True = show only stations with geo_info;
            False = show only stations without geo_info.

        ``has_extended_info`` : |str, Boolean|, optional
            Not set = display all;
            True = show only stations which do provide extended information;
            False = show only stations without extended information.

        ``is_https`` : |str, Boolean|, optional
            Not set = display all;
            True = show only stations which have https url;
            False = show only stations that do stream unencrypted with http.

        ``order`` : |str|, optional
            Name of the attribute the result list will be sorted by.

        ``reverse`` : |str, Boolean|, optional
            Reverse the result list if set to true.

        ``offset`` : |str|, optional
            Starting value of the result list from the database.
            For example, if you want to do paging on the server side.

        ``limit`` : |str|, optional
            Number of returned datarows (stations) starting with offset.

        ``hidebroken`` : |str, Boolean|, optional
            Do list/not list broken stations.
        """
        arguments = {
            "name": name,
            "nameExact": name_exact,
            "country": country,
            "countryExact": country_exact,
            "countrycode": countrycode,
            "state": state,
            "stateExact": state_exact,
            "language": language,
            "languageExact": language_exact,
            "tag": tag,
            "tagExact": tag_exact,
            "tagList": tag_list,
            "codec": codec,
            "bitrateMin": bitrate_min,
            "bitrateMax": bitrate_max,
            "has_geo_info": has_geo_info,
            "has_extended_info": has_extended_info,
            "is_https": is_https,
            "order": order,
            "reverse": reverse,
            "offset": offset,
            "limit": limit,
            "hidebroken": hidebroken,
        }
        params = {}

        for argument in arguments.items():
            item, value = argument

            if value != NOT_SET:
                params[item] = value

        return run(self.__get(
            type=list[Station],
            endpoint="stations/search",
            params=params
        ))




"""Internal RadioBrowser API wrapper module"""

from random import choice
from socket import (
    getaddrinfo,
    gethostbyaddr,
    gaierror,
    herror,
    IPPROTO_TCP,
)
from typing import Union

from ..datatypes import URL
from ..decorators.memoizations import memoize


class Browser:
    """Internal RadioBrowser API wrapper class

    Attributes:
    ----------------------------------------
    ``base`` : property
        Base Radio Browser API url

    ``servers`` : property
        Radio Browser API servers
    
    ``lookup`` : method
        Reverse DNS lookup

    ``hosts`` : property
        Radio Browser API hosts

    ``url`` : property
        Pick random Radio Browser API url
    """

    def __init__(self) -> None:
        """"""
        ...

    @property
    @memoize
    def base(self) -> str:
        """Base Radio Browser API url"""
        return "api.radio-browser.info"

    @property
    @memoize
    def servers(self) -> list[Union[str, URL]]:
        """Radio Browser API servers"""
        try:
            data = getaddrinfo(host=f"all.{self.base}",
                               port=80, family=0, type=0,
                               proto=IPPROTO_TCP)
        except gaierror:
            ...
        else:
            if (data and (isinstance(data[0][4], tuple))):
                return [ip[4][0] for ip in data]

        return []

    def lookup(self, ip: str) -> str:
        """Reverse DNS lookup"""
        try:
            hostname, *_ = gethostbyaddr(ip)
        except herror:
            ...

        return hostname

    @property
    @memoize
    def hosts(self) -> list[Union[str, URL]]:
        """Radio Browser API hosts"""
        names = []

        for ip in self.servers:
            try:
                host = self.lookup(ip)
            except Exception:
                ...
            else:
                names.append(host)

        return names

    @property
    @memoize
    def url(self) -> Union[str, URL]:
        """Pick random Radio Browser API url"""
        host = choice(sorted(self.hosts))

        return f"https://{host}"


"""RadioBrowser API Struct Language"""

from msgspec import Struct

from typing import Optional


class Language(Struct):
    """RadioBrowser API Struct Language

    Attributes
    ----------------------------------------
    ``name`` : |str|, optional
        Name of the language

    ``iso_639`` : |str|, optional
        Standards by the International Organization for Standardization

    ``stationcount`` : |int|, optional
        Station language count
    """
    name: Optional[str] = None
    iso_639: Optional[str] = None
    stationcount: Optional[int] = None


"""RadioBrowser API Struct Station Check"""

from typing import Optional, Union

from msgspec import Struct

from radiozilla.datatypes import UUID, URL


class StationCheck(Struct):
    """RadioBrowser API Struct Station Check

    Attributes:
    ----------------------------------------
    ``checkuuid`` : |str, UUID|, optional
        An unique id for this StationCheck.

    ``stationuuid`` : |str, UUID|, optional
        An unique id for referencing a Station.

    ``source`` : |str|, optional
        DNS Name of the server that did the stream check.

    ``codec`` : |str|, optional
        High level name of the used codec of the stream.
        May have the format AUDIO or AUDIO/VIDEO.

    ``bitrate`` : |int|, optional
        Bitrate 1000 bits per second (kBit/s) of the stream. (Audio + Video)

    ``hls`` : |int|, optional
        1 means this is an HLS stream, otherwise 0.

    ``ok`` : |int|, optional
        1 means this stream is reachable, otherwise 0.

    ``timestamp_iso8601`` : |str|, optional
        Date and time of check creation.

    ``timestamp`` : |str|, optional
        Date and time of check creation.

    ``urlcache`` : |str, URL|, optional
        Direct stream url that has been resolved from the
        main stream url. HTTP redirects and playlists
        have been decoded. If hls==1 then this is still a HLS-playlist.

    ``metainfo_overrides_database`` : |int|, optional
        1 means this stream has provided extended information and
        it should be used to override the local database, otherwise 0.

    ``public`` : |int|, optional
        1 that this stream appears in the public
        shoutcast/icecast directory, otherwise 0.

    ``name`` : |str|, optional
        The name extracted from the stream header.

    ``description`` : |str|, optional
        The description extracted from the stream header.

    ``tags`` : |str|, optional
        Komma separated list of tags. (genres of this stream)

    ``countrycode`` : |str|, optional
        Official countrycodes as in ISO 3166-1 alpha-2.

    ``countrysubdivisioncode`` : |str|, optional
        Official country subdivision codes as in ISO 3166-2.

    ``homepage`` : |str|, optional
        The homepage extracted from the stream header.

    ``favicon`` : |str|, optional
        The favicon extracted from the stream header.

    ``loadbalancer`` : |str|, optional
        The loadbalancer extracted from the stream header.

    ``server_software`` : |str|, optional
        The name of the server software used.

    ``sampling`` : |int|, optional
        Audio sampling frequency in Hz.

    ``timing_ms`` : |int|, optional
        Timespan in miliseconds this check needed to be finished.

    ``languagecodes`` : |str|, optional
        The description extracted from the stream header.

    ``ssl_error`` : |int|, optional
        1 means that a ssl error occured while
        connecting to the stream, 0 otherwise.

    ``geo_lat`` : |int|, optional
        Latitude on earth where the stream is located.

    ``geo_long`` : |int|, optional
        Longitude on earth where the stream is located.
    """

    checkuuid: Optional[Union[str, UUID]] = None
    stationuuid: Optional[Union[str, UUID]] = None
    source: Optional[str] = None
    codec: Optional[str] = None
    bitrate: Optional[int] = None
    hls: Optional[int] = None
    ok: Optional[int] = None
    timestamp_iso8601: Optional[str] = None
    timestamp: Optional[str] = None
    urlcache: Optional[Union[str, URL]] = None
    metainfo_overrides_database: Optional[int] = None
    public: Optional[int] = None
    name: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[str] = None
    countrycode: Optional[str] = None
    countrysubdivisioncode: Optional[str] = None
    homepage: Optional[str] = None
    favicon: Optional[str] = None
    loadbalancer: Optional[str] = None
    server_software: Optional[str] = None
    sampling: Optional[int] = None
    timing_ms: Optional[int] = None
    languagecodes: Optional[str] = None
    ssl_error: Optional[int] = None
    geo_lat: Optional[int] = None
    geo_long: Optional[int] = None


"""RadioBrowser API Struct Station"""

from typing import Optional, Union

from msgspec import Struct

from radiozilla.datatypes import UUID, URL


class Station(Struct):
    """RadioBrowser API Struct Station

    Attributes:
    ----------------------------------------
    ``changeuuid`` : |str, UUID|, optional
        A globally unique identifier for the change of the station information.

    ``stationuuid`` : |str, UUID|, optional
        A globally unique identifier for the station.

    ``name`` : |str|, optional
        The name of the station.

    ``url`` : |str, URL|, optional
        The stream URL provided by the user.

    ``url_resolved`` : |str, URL|, optional
        An automatically "resolved" stream URL. Things resolved are
        playlists (M3U/PLS/ASX...), HTTP redirects (Code 301/302).
        This link is especially usefull if you use this API from a
        platform that is not able to do a resolve on its own 
        (e.g. JavaScript in browser) or you just don't want to invest
        the time in decoding playlists yourself.

    ``homepage`` : |str, URL|, optional
        URL to the homepage of the stream, so you can direct
        the user to a page with more information about the stream.

    ``favicon`` : |str, URL|, optional
        URL to an icon or picture that represents the stream. (PNG, JPG)

    ``tags`` : |str|, optional
        Tags of the stream with more information about it.

    ``country`` : |str|, optional
        DEPRECATED: use countrycode instead, full name of the country.
        Currently it is autogenerated from the countrycode.

    ``countrycode`` : |str|, optional
        Official countrycodes as in ISO 3166-1 alpha-2.

    ``state`` |str|, optional
        Full name of the entity where the
        station is located inside the country.

    ``language`` |str|, optional
        Languages that are spoken in this stream.

    ``languagecodes`` |str|, optional
        Languages that are spoken in this stream by code ISO 639-2/B.

    ``votes`` |int|, optional
        Number of votes for this station. This number is by server
        and only ever increases. It will never be reset to 0.

    ``lastchangetime`` |str|, optional
        Last time when the stream information was changed in the database.

    ``lastchangetime_iso8601`` |str|, optional
        Last time when the stream information was changed in the database.

    ``codec`` |str|, optional
        The codec of this stream recorded at the last check.

    ``bitrate`` |int|, optional
        The bitrate of this stream recorded at the last check.

    ``hls`` |int|, optional
        Mark if this stream is using HLS distribution or non-HLS.

    ``lastcheckok`` |str|, optional
        The current online/offline state of this stream.
        This is a value calculated from multiple measure
        points in the internet. The test servers are located
        in different countries. It is a majority vote.

    ``lastchecktime`` |str|, optional
        The last time when any radio-browser server checked
        the online state of this stream.

    ``lastchecktime_iso8601`` |str|, optional
        The last time when any radio-browser server checke
        the online state of this stream.

    ``lastcheckoktime`` |str|, optional
        The last time when the stream was checked for the
        online status with a positive result.

    ``lastcheckoktime_iso8601`` |str|, optional
        The last time when the stream was checked for the
        online status with a positive result.

    ``lastlocalchecktime`` |str|, optional
        The last time when this server checked the online
        state and the metadata of this stream.

    ``lastlocalchecktime_iso860`` |str|, optional
        The last time when this server checked the online
        state and the metadata of this stream.

    ``clicktimestamp`` |str|, optional The time of the last click recorded for this stream

    ``clicktimestamp_iso8601`` |str|, optional
        The time of the last click recorded for this stream

    ``clickcount`` |int|, optional
        Clicks within the last 24 hours

    ``clicktrend`` |int|, optional
        The difference of the clickcounts within the last 2 days.
        Posivite values mean an increase, negative a decrease of clicks.

    ``ssl_error`` |int|, optional
        0 means no error, 1 means that there was an ssl error while connecting to the stream url.

    ``geo_lat`` |int|, optional
        Latitude on earth where the stream is located.

    ``geo_long`` |int|, optional
        Longitude on earth where the stream is located.

    ``has_extended_info`` |bool|, optional
        Is true, if the stream owner does provide extended
        information as HTTP headers which override the
        information in the database.
    """

    changeuuid: Optional[Union[str, UUID]] = None
    stationuuid: Optional[Union[str, UUID]] = None
    name: Optional[str] = None
    url: Optional[Union[str, URL]] = None
    url_resolved: Optional[Union[str, URL]] = None
    homepage: Optional[Union[str, URL]] = None
    favicon: Optional[Union[str, URL]] = None
    tags: Optional[str] = None
    country: Optional[str] = None
    countrycode: Optional[str] = None
    state: Optional[str] = None
    language: Optional[str] = None
    languagecodes: Optional[str] = None
    votes: Optional[int] = None
    lastchangetime: Optional[str] = None
    lastchangetime_iso8601: Optional[str] = None
    codec: Optional[str] = None
    bitrate: Optional[int] = None
    hls: Optional[int] = None
    lastcheckok: Optional[int] = None
    lastchecktime: Optional[str] = None
    lastchecktime_iso8601: Optional[str] = None
    lastcheckoktime: Optional[str] = None
    lastcheckoktime_iso8601: Optional[str] = None
    lastlocalchecktime: Optional[str] = None
    lastlocalchecktime_iso860: Optional[str] = None
    clicktimestamp: Optional[str] = None
    clicktimestamp_iso8601: Optional[str] = None
    clickcount: Optional[int] = None
    clicktrend: Optional[int] = None
    ssl_error: Optional[int] = None
    geo_lat: Optional[Union[int, float]] = None
    geo_long: Optional[Union[int, float]] = None
    has_extended_info: Optional[bool] = None


"""RadioBrowser API Struct Country"""

from msgspec import Struct

from typing import Optional


class Country(Struct):
    """RadioBrowser API Struct Country

    Attributes
    ----------------------------------------
    ``name`` : |str|, optional
        Name of the country

    ``iso_3166_1`` : |str|, optional
        Standard defining codes 

    ``stationcount`` : |int|, optional
        Station country count
    """
    name: Optional[str] = None
    iso_3166_1: Optional[str] = None
    stationcount: Optional[int] = None


"""RadioBrowser API Datatypes"""

from typing import NewType, TypeVar


NOT_SET = ""

UUID = NewType("UUID", str)
URL = NewType("URL", str)
T = TypeVar("T")


class Order:
    """Posible values of Order argument"""

    NAME = "name"
    URL = "url"
    HOMEPAGE = "homepage"
    FAVICON = "favicon"
    TAGS = "tags"
    COUNTRY = "country"
    STATE = "state"
    LANGUAGE = "language"
    VOTES = "votes"
    CODEC = "codec"
    BITRATE = "bitrate"
    LASTCHECKOK = "lastcheckok"
    LASTCHECKTIME = "lastchecktime"
    CLICKCOUNT = "clickcount"
    CLICKTREND = "clicktrend"
    CHANGETIMESTAMP = "changetimestamp"
    RANDOM = "random"


class Boolean:
    """RadioBrowser API Boolean datatype"""

    BOTH = "both"
    TRUE = "true"
    FALSE = "false"


"""RadioBrowser API Struct State"""

from msgspec import Struct

from typing import Optional


class State(Struct):
    """RadioBrowser API Struct State

    Attributes
    ----------------------------------------
    ``name`` : |str|, optional
        Name of the state

    ``country`` : |str|, optional
        Country of the state

    ``stationcount`` : |int|, optional
        Station state count
    """
    name: Optional[str] = None
    country: Optional[str] = None
    stationcount: Optional[int] = None


"""Defaults values of arguments"""

from .datatypes import Order, Boolean


_default_order = Order.NAME
_default_reverse = Boolean.FALSE
_default_hidebroken = Boolean.FALSE
_default_offset = 0
_default_limit = 100000

_default_name_exact = Boolean.FALSE
_default_country_exact = Boolean.FALSE
_default_state_exact = Boolean.FALSE
_default_language_exact = Boolean.FALSE
_default_tag_exact = Boolean.FALSE
_default_has_geo_info = Boolean.FALSE
_default_has_extended_info = Boolean.FALSE
_default_is_https = Boolean.FALSE


"""Memoization module for optimizing requests"""

from typing import Callable, Any


def memoize(callback: Callable) -> Callable:
    """"""
    cache = {}

    def wrapper(*args, **kwargs) -> Any:
        item = str(args) + str(kwargs)

        if item not in cache:
            cache[item] = callback(*args, **kwargs)

        return cache[item]

    return wrapper